@@ -1,5 +1,5 @@
 /* Build executable statement trees.
-   Copyright (C) 2000, 2001, 2002, 2004, 2005, 2006, 2007
+   Copyright (C) 2000, 2001, 2002, 2004, 2005, 2006, 2007, 2008
    Free Software Foundation, Inc.
    Contributed by Andy Vaught
 
@@ -36,7 +36,6 @@ gfc_code new_st;
 void
 gfc_clear_new_st (void)
 {
-
   memset (&new_st, '\0', sizeof (new_st));
   new_st.op = EXEC_NOP;
 }
@@ -49,7 +48,7 @@ gfc_get_code (void)
 {
   gfc_code *c;
 
-  c = gfc_getmem (sizeof (gfc_code));
+  c = XCNEW (gfc_code);
   c->loc = gfc_current_locus;
   return c;
 }
@@ -59,30 +58,28 @@ gfc_get_code (void)
    its tail, returning a pointer to the new tail.  */
 
 gfc_code *
-gfc_append_code (gfc_code * tail, gfc_code * new)
+gfc_append_code (gfc_code *tail, gfc_code *new_code)
 {
-
   if (tail != NULL)
     {
       while (tail->next != NULL)
 	tail = tail->next;
 
-      tail->next = new;
+      tail->next = new_code;
     }
 
-  while (new->next != NULL)
-    new = new->next;
+  while (new_code->next != NULL)
+    new_code = new_code->next;
 
-  return new;
+  return new_code;
 }
 
 
 /* Free a single code structure, but not the actual structure itself.  */
 
 void
-gfc_free_statement (gfc_code * p)
+gfc_free_statement (gfc_code *p)
 {
-
   if (p->expr)
     gfc_free_expr (p->expr);
   if (p->expr2)
@@ -111,6 +108,7 @@ gfc_free_statement (gfc_code * p)
     case EXEC_ARITHMETIC_IF:
       break;
 
+    case EXEC_COMPCALL:
     case EXEC_CALL:
     case EXEC_ASSIGN_CALL:
       gfc_free_actual_arglist (p->ext.actual);
@@ -149,6 +147,10 @@ gfc_free_statement (gfc_code * p)
       gfc_free_inquire (p->ext.inquire);
       break;
 
+    case EXEC_WAIT:
+      gfc_free_wait (p->ext.wait);
+      break;
+
     case EXEC_READ:
     case EXEC_WRITE:
       gfc_free_dt (p->ext.dt);
@@ -156,7 +158,7 @@ gfc_free_statement (gfc_code * p)
 
     case EXEC_DT_END:
       /* The ext.dt member is a duplicate pointer and doesn't need to
-         be freed.  */
+	 be freed.  */
       break;
 
     case EXEC_FORALL:
@@ -170,13 +172,14 @@ gfc_free_statement (gfc_code * p)
     case EXEC_OMP_PARALLEL_SECTIONS:
     case EXEC_OMP_SECTIONS:
     case EXEC_OMP_SINGLE:
+    case EXEC_OMP_TASK:
     case EXEC_OMP_WORKSHARE:
     case EXEC_OMP_PARALLEL_WORKSHARE:
       gfc_free_omp_clauses (p->ext.omp_clauses);
       break;
 
     case EXEC_OMP_CRITICAL:
-      gfc_free ((char *) p->ext.omp_name);
+      gfc_free (CONST_CAST (char *, p->ext.omp_name));
       break;
 
     case EXEC_OMP_FLUSH:
@@ -188,6 +191,7 @@ gfc_free_statement (gfc_code * p)
     case EXEC_OMP_MASTER:
     case EXEC_OMP_ORDERED:
     case EXEC_OMP_END_NOWAIT:
+    case EXEC_OMP_TASKWAIT:
       break;
 
     default:
@@ -199,7 +203,7 @@ gfc_free_statement (gfc_code * p)
 /* Free a code statement and all other code structures linked to it.  */
 
 void
-gfc_free_statements (gfc_code * p)
+gfc_free_statements (gfc_code *p)
 {
   gfc_code *q;
 
